from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING: from cord import Cord

from PIL import Image
from PIL import ImageDraw

from ..vector2 import Vector2
from ..colors import *
from ..font import Font

class Component:
    def __init__(_, cord:Cord=None, x:int=0, y:int=0, parent:"Component"=None,
                 width:int|None=0, height:int|None=0,
                 color:Color=None,background:Color=GRAY, font:Font=None,
                 border:bool=False):
        _.cord:Cord = cord
        _.parent:Component|None = parent
        _.color:Color = color
        _.background:Color = background
        _.border:bool = border
        _.border_color:Color = WHITE
        _.border_width:int = 1
        _.children:list[Component] = []
        _.font:Font = font if font else (parent.font if parent else _.cord.font)
        _.x:int = x
        _.y:int = y
        _.width:int = _.cord.width if width is None else width
        _.height:int = _.cord.height if height is None else height
        _.image_width:int = _.cord.width if not _.parent else _.parent.width
        _.image_height:int = _.cord.height if not _.parent else _.parent.height
        _._determine_dimensions()
        _.path:str = _.parent.path + f".{_.__class__.__name__}" if parent else _.__class__.__name__
        _.image:Image = None
        _.drawing:ImageDraw = None


    @property
    def x_center(_) -> int: return _.width // 2
    @property
    def y_center(_) -> int: return _.height // 2
    @property
    def image_center(_) -> int: return Vector2(_.x_center, _.y_center)


    def __str__(_) -> str: return _.path


    def _determine_dimensions(_) -> None:
        if _.parent:
            if _.parent.border:
                _.x = _.parent.x + _.x + _.parent.border_width
                _.y = _.parent.y + _.y + _.parent.border_width
                _.width = _.parent.width - _.parent.border_width * 2
                _.height = _.parent.height - _.parent.border_width * 2
            else:
                _.x = _.x + _.parent.x
                _.y = _.y + _.parent.y
                _.width = _.parent.width
                _.height = _.parent.height


    async def debug(_, vertical_center:bool=False, horizontal_center:bool=False) -> None:
        if vertical_center:
            await _.cord.line(parent=_, start=Vector2(_.x_center, 0), end=Vector2(_.x_center, _.height), fill_width=3, color=DEBUG_COLOR)
        if horizontal_center:
            await _.cord.line(parent=_, start=Vector2(0, _.y_center), end=Vector2(_.width, _.y_center), fill_width=3, color=DEBUG_COLOR)


    async def draw(_) -> None:
        # print(f"Drawing Base of {_.__str__}")
        _.image = Image.new("RGBA", (_.width, _.height), color=_.background)
        _.drawing = ImageDraw.Draw(_.image)
        await _.construct_components()


    async def construct_components(_) -> None:
        # print(f"Constructing {_}'s Components")
        child:Component
        for child in _.children:
            child_image = await child.draw()
            _.image.paste(child_image, (child.x, child.y), mask=child_image.split()[3])


    async def get_text_width(_, text:str, font:Font=None) -> int:
        test_font:Font = font if font is not None else _.cord.font
        measuring_image = Image.new("RGBA", (10, 10))
        drawing = ImageDraw.Draw(measuring_image)
        return int(drawing.textlength(text, font=test_font.font))



async def format_numeric(numeric: int | float) -> str:
    suffixes = [
        (1_000_000_000_000_000, "Q"),
        (1_000_000_000_000, "T"),
        (1_000_000_000, "B"),
        (1_000_000, "M"),
        (1_000, "K"),
    ]
    abs_value = abs(numeric)

    for threshold, suffix in suffixes:
        if abs_value >= threshold:
            value = numeric / threshold
            return f"{value:.0f}{suffix}" if value.is_integer() else f"{value:.1f}{suffix}"
    return str(numeric)


# Generalized imports for dev utility
from PIL import Image
from discord import Interaction, Message
from discord.ext.commands import Context
from discord.abc import GuildChannel
from asyncio import sleep, create_task

# CordForge specific
from .cord import Cord
from .card import Card
from .object import Object
from .vector2 import Vector2
from .colors import *
from .font import Font
from .components import *
from .utilities import *
from .user import User

from sys import stdout

import logging

logging.basicConfig(stream=stdout, level=logging.INFO)
logging.basicConfig(stream=stdout, level=logging.INFO)


from PIL import Image as Image
from ..font import Font

class BoardItem:
    def __init__(_, text:str, image:Image=None, separation:int=4, font:Font=None):
        _.image = image
        _.text = text
        _.font = font if font is not None else None
        _.separation = separation


from sys import exit, platform, executable
from os import remove, getcwd
from os.path import join
from subprocess import  Popen, PIPE, STDOUT
from glob import glob
from sys import argv, stdout
from asyncio import run as async_run

import tkinter as tk
from tkinter.scrolledtext import ScrolledText
import threading
import logging

class LoggingHandler(logging.Handler):
    def __init__(_, launcher):
        super().__init__()
        _.launcher = launcher

    def emit(_, record):
        try:
            msg = _.format(record)
            _.launcher.append_text(msg)
        except Exception:
            _.handleError(record)


class Launcher:
    def __init__(_):
        logging.basicConfig(
            level=logging.INFO,
            stream=stdout,
            format="%(levelname)s:%(name)s:%(message)s"
        )
        _.logger = logging.getLogger("Launcher")
        handler = LoggingHandler(_)
        handler.setFormatter(logging.Formatter("%(levelname)s:%(name)s:%(message)s"))
        logging.getLogger().addHandler(handler)

        _.bot = None
        _.key = None
        _.working_directory = getcwd()
        _.commands = {"start": _.start,
                      "restart": _.restart,
                      "exit": _.exit,
                      "stop": _.stop,
                      "//": _.emergency_stop,
                      "clear logs": _.clear_logs}
        
        if len(argv) == 2:
            _.key_selection = argv[1]
        else:
            _.logger.info("No key chosen, finding first in Keys file.")
            with open(join(_.working_directory, "keys"), 'r') as keys:
                _.key_selection = keys.readlines()[0].split("=")[0].strip()

        _.python = executable
        _.call_command = [_.python, "-u", "-B", "entry.py", _.key_selection]

        _.construct_window()


    def construct_window(_) -> None:
        _.root = tk.Tk()
        _.root.title(f"{_.working_directory}")

        _.root.protocol("WM_DELETE_WINDOW", _.close_window)

        _.frame = tk.Frame(_.root, bg="grey")
        _.frame.pack(fill="both", padx=5, pady=5)

        _.text = ScrolledText(_.frame, bg="#e2e2e2", state="disabled", wrap="word")
        _.text.pack(fill="both", expand=True)

        _.entry = tk.Entry(_.frame, bg="lightgrey")
        _.entry.pack(fill="both", expand=True)
        _.entry.bind("<Return>", _.send_input)
        _.root.mainloop()


    def close_window(_) -> None:
        if _.bot:
            print("Bot is running, killing before closing")
            _.bot.kill()
        _.root.quit()


    def read_stream(_, stream, tag):
        for line in stream:
            _.append_text(f"[{tag}] {line}")


    def append_text(_, msg:str):
        raw_message = msg[9:]
        if msg.startswith("[stdout]"): msg = raw_message
        
        if not msg.endswith("\n"): msg += "\n"

        _.text.configure(state="normal")
        _.text.insert("end", msg)
        _.text.see("end")
        _.text.configure(state="disabled")


    def send_input(_, event):
        user_input = _.entry.get()
        _.append_text(f"Input: {user_input}\n")
        if user_input in _.commands.keys():
            _.commands[user_input]()
        elif _.bot:
            _.bot.stdin.write(user_input + "\n")
            _.bot.stdin.flush()
        _.entry.delete(0, "end")


    def start(_):
        _.logger.info("Starting Bot...")
        _.bot = Popen(_.call_command, stdin=PIPE, stdout=PIPE, stderr=STDOUT, text=True, bufsize=1)

        threading.Thread(target=_.read_stream, args=(_.bot.stdout, "stdout"), daemon=True).start()
        threading.Thread(target=_.read_stream, args=(_.bot.stderr, "stderr"), daemon=True).start()


    def restart(_):
        if _.bot:
            _.logger.info("Restarting Discord bot...")
            _.bot = _.bot.kill()
            _.start()
            _.logger.info("Discord bot restarted")
        else:
            _.logger.info("There isn't a running bot")


    def exit(_):
        if not _.bot:
            exit()
        else:
            _.logger.info("There is a running bot")


    def stop(_):
        if _.bot:
            _.logger.info("Discord bot stopped")
            _.bot = _.bot.kill()
        else:
            print("There isn't a running bot")


    def emergency_stop(_):
        if not _.bot:
            _.logger.info("Bot is not running it seems, stopping altogether though.")
            exit()

        if _.bot:
            _.logger.info("Discord bot stopped")
            _.bot = _.bot.kill()
            exit()


    def clear_logs(_):
        for file in glob("Source\\Logs\\*.log"):
            try:
                remove(file)
            except OSError:
                print("Error removing log files for some reason")


if __name__ == "__main__":
    Launcher()